/***************/
/*  TableInit  */
/***************/
TableInit: PROC( TABLE() VARENTRY IDENT, TXTList() CHAR(10) IDENT, Nxttxt FIXED(31) IDENT ) ;
  /* Die leeren Pltze der Tabelle werden neu initialisiert */
  DCL Empty BIT INIT('0'B);
  FOR i TO 1 UPB TABLE REPEAT
    IF Empty
    THEN
      TABLE(i) = EmptySlot ;
    ELSE
      TABLE(i) = VARTABINIT(i);
      Empty = TABLE(i).Descr == vtUnused;
    FIN;
  END;
  Empty = '0'B;
  FOR i TO 1 UPB TXTList REPEAT
    IF Empty
    THEN
      TXTList(i) = ' ';
    ELSE
      TXTList(i) = TxtPoolInit(i);
      Empty = TXTList(i) == ' ';
      Nxttxt = i;
    FIN;
  END;
END ; /* PROC TableInit -------------------------------------------*/

/*
! *******************************************************************
!   BASE64 converts in-place to Base64-encoding                     *
! *******************************************************************
BASE64: PROC( SRC CHAR(1) IDENT, Len FIXED, MaxNewLen FIXED ) RETURNS( FIXED );

DCL B24 BIT(32);
DCL PB24 REF CHAR(3);  ! points to B24
DCL PEnc REF CHAR(4);
DCL PSrc REF CHAR(3);
DCL VoidPtr REF STRUCT[];

   IF ( Len + 2 ) // 3 * 4 > MaxNewLen
   THEN
     RETURN(0);
   FIN;

   ! PEnc to end of encrypted string
   VoidPtr = SRC;
   PEnc = VoidPtr;
   CALL REFADD( PEnc, Len // 3 );

   ! PSrc to end of plain string
   VoidPtr = SRC;
   PSrc = VoidPtr;
   CALL REFADD( PSrc, Len // 3 );

   ! PB24 to B24
   VoidPtr = B24;
   PB24 = VoidPtr;
   B24 = '00000000'B4;

   ! convert trailing bytes
   CASE Len REM 3
   ALT (0)  ! Laenge durch 3 teilbar
     CALL REFADD( PSrc, (-1) );
     CALL REFADD( PEnc, (-1) );

   ALT (1)  ! 2 Bytes fehlen
     PB24.CHAR(1) = PSrc.CHAR(1);
     CALL REFADD( PSrc, (-1) );

     PEnc.CHAR(1) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-26) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(2) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-20) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(3) = '=';
     PEnc.CHAR(4) = '=';
     CALL REFADD(PEnc, (-1) );

   ALT (2)  ! 1 Byte fehlt
     PB24.CHAR(1) = PSrc.CHAR(1);
     PB24.CHAR(2) = PSrc.CHAR(2);
     CALL REFADD( PSrc, (-1) );

     PEnc.CHAR(1) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-26) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(2) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-20) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(3) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-14) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(4) = '=';
     CALL REFADD(PEnc, (-1) );

   FIN;

   ! loop over plain string
   TO Len // 3 REPEAT
     CONT PB24 = PSrc;
     CALL REFADD( PSrc, (-1) );

     PEnc.CHAR(1) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-26) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(2) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-20) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(3) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (-14) ) AND '0000003F'B4 ) + 1 );
     PEnc.CHAR(4) = Base64Encoding.CHAR( TOFIXED( ( B24 SHIFT (- 8) ) AND '0000003F'B4 ) + 1 );
     CALL REFADD(PEnc, (-1) );
   END;

   RETURN( ( Len + 2 ) // 3 * 4 );
END;
*/

! *******************************************************************
!   NUMBER checks whether the CHAR(1) obj is a dec. digit           *
! *******************************************************************
NUMBER: PROC(PTR REF CHAR(1)) RETURNS(BIT(1));
   CASE PTR
     ALT('0':'9') RETURN('1'B);
     OUT          RETURN('0'B);
   FIN;
END;

! *******************************************************************
!   LETTER checks whether the CHAR(1) obj is a letter or underscore *
! *******************************************************************
LETTER: PROC(PTR REF CHAR(1)) RETURNS(BIT(1));
   CASE PTR
     ALT ('A':'Z') RETURN('1'B);
     ALT ('a':'z') RETURN('1'B);
     ALT ('_')     RETURN('1'B);
     OUT RETURN('0'B);
   FIN;
END;

! *******************************************************************
!   GETINT delivers an integer from the Text by REF CHAR(1)         *
! *******************************************************************
GETINT: PROC(PTR REF CHAR(1) IDENT) RETURNS(FIXED(31));

  DCL BUF FIXED(31);
  DCL CPTR REF CHAR(1);

  BUF=0(31);
  CPTR=PTR;

  REPEAT
    BUF=BUF*10+(TOFIXED(CPTR)-48);
    REFADD(CPTR,1);
    IF NOT NUMBER(CPTR) THEN
      PTR=CPTR; RETURN(BUF);
    FIN;
  END;
END;

! *******************************************************************
! RUBBL Rubbers all Blanks moving the pointer
! *******************************************************************
RUBBL: PROC(PTR REF CHAR(1) IDENT);
  DCL CPTR REF CHAR(1);

  CPTR=PTR;
  REPEAT;
    IF CPTR NE ''\20\'' THEN PTR=CPTR; RETURN; FIN;
    REFADD(CPTR,1);
  END;
END;

! *******************************************************************
! IDSCAN processes an max 24 char long identifier
! *******************************************************************
IDSCAN: PROC(PTR REF CHAR(1) IDENT, IDLen FIXED IDENT) RETURNS(CHAR(24));
  DCL RESULT CHAR(24);
  DCL CPTR2 REF CHAR(1);

  RESULT=''; IDLen=0; CPTR2=RESULT.CHAR(1);
  WHILE (LETTER(PTR) OR NUMBER(PTR) OR PTR =='.' OR PTR =='/' OR PTR == '$') REPEAT
      CONT CPTR2 = PTR; IDLen=IDLen+1;
      REFADD(PTR,1); REFADD(CPTR2,1);
      IF IDLen GE 24 THEN EXIT; FIN;
  END;    ! Identifier-scan
  RETURN(RESULT);
END; ! IDSCAN

! *******************************************************************
! SERID searches a max 24 char long identifier in the list
! *******************************************************************
SERID: PROC(TABLE() VARENTRY IDENT,IDSTR CHAR(24) IDENT,DESCR FIXED) RETURNS(FIXED);
  FOR SERINDEX TO 1 UPB TABLE REPEAT
    IF     ( TABLE(SERINDEX).Name == IDSTR )
       AND ( TABLE(SERINDEX).Descr == DESCR OR DESCR == vtIgnore )
    THEN
      RETURN(SERINDEX FIT 1);
    FIN;
  END;
  RETURN(0);
END;

! *******************************************************************
!    INIREAD reads the given INI-File and assigns the values
! *******************************************************************
INIREAD: PROC(FILEX DATION IN ALPHIC IDENT,
              TABLE() VARENTRY IDENT,
              TXTList() CHAR(10) IDENT,
              Filename CHAR(255) IDENT,
              NextTX FIXED(31) IDENT)
              RETURNS(FIXED(31));

  DCL BUFLINE CHAR(255);                             ! Buffers one line
  DCL CC FIXED;                                      ! Char counter
  DCL (CPTR,CPTR2,CPTR3) REF CHAR(1);                ! Text pointer
  DCL IDENTIFIER   CHAR(24);
  DCL SERINDEX     FIXED;                            ! For searching in
  DCL LINENO       FIXED(31);                        ! Line cntr
  DCL (ITEMP,ITEMP2) FIXED(31);                      ! Temp Integer
  DCL TABREF REF VARENTRY;
  DCL IDLEN FIXED;                                   ! LEN of ID
  DCL (Charcnt,Charcntnb) FIXED(31);
  DCL (Itemp,Itemp2)  FIXED;
  DCL (C1BUF,C1BUF2) CHAR(1);
  DCL VALLINE CHAR(255);
  DCL ENVLINE CHAR(255);

  OPEN FILEX BY IDF(Filename),OLD;
  CPTR=BUFLINE.CHAR(1);
  BUFLINE='Cannot open file HTTPSERV.INI.';
  LINENO=0(31);
  IF ST(FILEX) NE 0 THEN GOTO READERR; FIN;
  REWIND(FILEX);

  WHILE ST(FILEX) == 0 REPEAT
    GET BUFLINE FROM FILEX BY SKIP,A;                ! Read one line
    LINENO=LINENO+1(31);                             ! Move counter
    IF LEN(BUFLINE) LT 255 THEN
       BUFLINE.CHAR(LEN(BUFLINE)+1)=''\0D\'';
    ELSE
       BUFLINE.CHAR(255) = ''\0D\'';                 ! set a CR
    FIN;
    CPTR=BUFLINE.CHAR(1);
    IF CPTR == ''\0A\''
    THEN
      CALL REFADD( CPTR, 1 );                        ! MS-DOS-File
    FIN;

    RUBBL(CPTR);                                     ! Skip leading Bl
! .... We are at the real beginning of the line .....................

    IF CPTR NE '!' THEN                              ! Skip all if !

       IDENTIFIER=IDSCAN(CPTR,IDLEN);                ! Fetch the ID
       IF IDLEN == 0 THEN GOTO SYNFAIL; FIN;
       Itemp=SERID(TABLE,IDENTIFIER,vtIgnore);
       IF Itemp == 0 THEN GOTO UNDEF; FIN;
       TABREF=TABLE(Itemp);

       SWITCH:

       CASE TABREF.Descr
         ! ... Get an integer number
         ALT (vtNumVal)
                  RUBBL(CPTR);                            ! Rubber Blank
                  IF CPTR NE '=' THEN GOTO SYNFAIL; FIN;  ! Syntax error?
                  REFADD(CPTR,1);                         ! Skip =
                  RUBBL(CPTR);                            ! Rubber Blank
                  IF NOT NUMBER(CPTR) THEN GOTO SYNFAIL;FIN;
                  TABREF.IValue=GETINT(CPTR);
         ! ... Get a piece of Text until end of line
         ALT (vtTextStr,vtFileType,vtApplication,vtApplication2,vtApplication3,vtDirAlias,vtRealm)
                  RUBBL(CPTR);                        ! Rubber Blank
                  IF CPTR NE '=' THEN GOTO SYNFAIL; FIN;  ! Syntax error?
                  REFADD(CPTR,1);                         ! Skip =
                  RUBBL(CPTR);                            ! Rubber Blank
                  IF NextTX GT 1 UPB TXTList THEN
                    GOTO OVERR;
                  FIN;                          ! Possible Text-Overflow
                  TABREF.IValue=NextTX;
                  VALLINE=' ';                            ! copy to VALLINE
                  CPTR2=VALLINE.CHAR(1);
                  Charcnt=0(31); Charcntnb=0(31);
                  WHILE CPTR NE ''\0D\'' REPEAT
                    Charcnt=Charcnt+1(31);
                    IF CPTR NE ' ' THEN Charcntnb=Charcnt; FIN;
                    CONT CPTR2=CPTR; REFADD(CPTR,1); REFADD(CPTR2,1);
                    IF Charcntnb > 255 THEN
                      GOTO OVERR;
                    FIN;
                  END;
                  ENVLINE=ENVGET(VALLINE);                ! replace environment variables
                  IF ENVLINE.CHAR(1)=='$' THEN
                     ENVLINE = VALLINE;                   ! restore original text
                  FIN;
                  CPTR3=ENVLINE.CHAR(1);                  ! copy to text-pool
                  CPTR2=TXTList(NextTX).CHAR(1);
                  Charcnt=0(31); Charcntnb=0(31);
                  TO LEN(ENVLINE) WHILE CPTR3 NE ''\0D\'' REPEAT
                    Charcnt=Charcnt+1(31);
                    IF CPTR3 NE ' ' THEN Charcntnb=Charcnt; FIN;
                    CONT CPTR2=CPTR3; REFADD(CPTR3,1); REFADD(CPTR2,1);
                    IF NextTX+(Charcntnb+9(31))//10(31) > 1 UPB TXTList THEN
                          GOTO OVERR;
                    FIN;
                  END;
                  NextTX=NextTX+(Charcntnb+9(31))//10(31);
                  TABREF.Len=Charcntnb FIT 1;
         ! ... Get an Authenticate definition .....................
         ALT (vtAuthenticate)
                  RUBBL(CPTR);                        ! Rubber Blank
                  IF CPTR NE '=' THEN GOTO SYNFAIL; FIN;  ! Syntax error?
                  REFADD(CPTR,1);                         ! Skip =
                  RUBBL(CPTR);                            ! Rubber Blank
                  IF NextTX GT 1 UPB TXTList THEN
                    GOTO OVERR;
                  FIN;                          ! Possible Text-Overflow
                  TABREF.IValue=NextTX;;
                  CPTR2=TXTList(NextTX).CHAR(1);
                  Charcnt=0(31); Charcntnb=0(31);
                  WHILE CPTR NE ''\0D\'' REPEAT
                    Charcnt=Charcnt+1(31);
                    IF CPTR NE ' ' THEN Charcntnb=Charcnt; FIN;
                    CONT CPTR2=CPTR; REFADD(CPTR,1); REFADD(CPTR2,1);
                    IF NextTX+(Charcntnb+9(31))//10(31) > 1 UPB TXTList THEN
                          GOTO OVERR;
                    FIN;
                  END;
                  NextTX=NextTX+(Charcntnb+9(31))//10(31);
                  TABREF.Len=Charcntnb FIT 1;
         ! ... Get an IP-Address or Mask .....................
         ALT (vtIPAddr)
                  RUBBL(CPTR);                            ! Rubber Blank
                  IF CPTR NE '=' THEN GOTO SYNFAIL; FIN;  ! Syntax error?
                  REFADD(CPTR,1);                         ! Skip =
                  RUBBL(CPTR);                            ! Rubber Blank
                  ITEMP=0(31);
                  FOR II TO 4 REPEAT
                     IF NOT NUMBER(CPTR) THEN GOTO SYNFAIL; FIN;
                     ITEMP2=GETINT(CPTR);
                     IF (ITEMP2 GT 255) THEN GOTO LIMERR; FIN;
                     IF II LT 4 THEN
                       IF (CPTR NE '.') THEN GOTO SYNFAIL; FIN;
                       REFADD(CPTR,1);
                     FIN;
                     ITEMP=ITEMP*256+ITEMP2;
                  END;  ! FOR II TO 4 REPEAT
                  TABREF.IValue=ITEMP;
         ! ... Define a new file-type, Application, DirAlias or Realm .....
         ALT (vtNewFileType,vtNewApplication,vtNewApplication2,vtNewApplication3,vtNewDirAlias,vtNewRealm)
                  CASE TABREF.Descr
                  ALT (vtNewFileType) Itemp2=vtFileType;
                  ALT (vtNewApplication) Itemp2=vtApplication;
                  ALT (vtNewApplication2) Itemp2=vtApplication2;
                  ALT (vtNewApplication3) Itemp2=vtApplication3;
                  ALT (vtNewDirAlias) Itemp2=vtDirAlias;
                  ALT (vtNewRealm) Itemp2=vtRealm;
                  ALT (vtNewAuthenticate) Itemp2=vtAuthenticate;
                  OUT      GOTO INTERR;
                  FIN;
                  RUBBL(CPTR);
                  IDENTIFIER=IDSCAN(CPTR,IDLEN);
                  IF IDLEN==0 THEN GOTO SYNFAIL; FIN;
                  Itemp=SERID(TABLE,IDENTIFIER,Itemp2);  ! ID really new?
                  IF Itemp NE 0 THEN GOTO DOUBDEF;FIN;
                  Itemp=SERID(TABLE,EMPTY,vtUnused);     ! Search a slot
                  IF Itemp == 0 THEN GOTO ALLFULL; FIN;
                  TABREF=TABLE(Itemp);    ! treat this new entry
                  TABREF.Name=IDENTIFIER; ! Insert ID
                  TABREF.Descr=Itemp2;    ! New mode
                  RUBBL(CPTR);
                  IF CPTR == '=' THEN GOTO SWITCH; FIN;
         ! ... Define a new Authenticate and check for existing realm .....
         ALT (vtNewAuthenticate)
                  RUBBL(CPTR);
                  IDENTIFIER=IDSCAN(CPTR,IDLEN);
                  IF IDLEN==0 THEN GOTO SYNFAIL; FIN;
                  Itemp=SERID(TABLE,IDENTIFIER,vtRealm);   ! realm-ID already known?
                  IF Itemp EQ 0 THEN GOTO UNDEF;FIN;
                  Itemp2=SERID(TABLE,EMPTY,vtUnused);       ! Search a slot
                  IF Itemp2 == 0 THEN GOTO ALLFULL; FIN;
                  TABREF=TABLE(Itemp2);    ! treat this new entry
                  TABREF.Name=' '; ! Insert index of realm
                  CONVERT Itemp TO TABREF.Name;
                  TABREF.Descr=vtAuthenticate;    ! New mode
                  RUBBL(CPTR);
                  IF CPTR == '=' THEN GOTO SWITCH; FIN;

         OUT      GOTO INTERR;

       FIN; ! Case type of object

       RUBBL(CPTR);
       IF CPTR NE '!' AND CPTR NE ''\0D\'' THEN GOTO SYNFAIL; FIN;

    FIN;  ! If CPTR NE '!'

  END;  ! While ST(..) = 0 Repeat

! ... The reading ended, may be end of file (which is ok here) ........

  IF ST(FILEX) == 1 THEN
    CLOSE FILEX; RETURN(LINENO); ! All is fine
  ELSE
    GOTO READERR;                ! Something wrong with file
  FIN;

LIMERR:  LINENO=LINENO+80000; GOTO FEXIT;
ALLFULL: LINENO=LINENO+70000; GOTO FEXIT;      ! All places full
DOUBDEF: LINENO=LINENO+60000; GOTO FEXIT;      ! Double defined
INTERR:  LINENO=LINENO+50000; GOTO FEXIT;      ! Internal error
OVERR:   LINENO=LINENO+40000; GOTO FEXIT;      ! Textbuffer overflow
READERR: LINENO=LINENO+30000; GOTO FEXIT;      ! Error reading INIT-File
SYNFAIL: LINENO=LINENO+20000; GOTO FEXIT;      ! Syntax error
UNDEF:   LINENO=LINENO+10000; GOTO FEXIT;      ! Symbol undefined

FEXIT:   CLOSE FILEX;
         CPTR2=BUFLINE.CHAR(255);
         C1BUF=CPTR;
         CONT CPTR=''\0A\'';
         WHILE CPTR ISNT CPTR2 REPEAT
           REFADD(CPTR,1);
           C1BUF2=CPTR;
           CONT CPTR=C1BUF;
           C1BUF=C1BUF2;
         END;
         Filename=BUFLINE;
         RETURN(LINENO);

END;   ! End the PROC
