! *********************************************************************
!      This procedure removes the addressed list from memory
! *********************************************************************

RM_LIST: PROC /*+G*/ (ROOT REF LISTE IDENT,loglink DATION OUT ALPHIC IDENT) GLOBAL;

 DCL (Tofree,Nxtblock) REF LISTE;

 #IF MAXILOG;
   DCL RMCOUNT FIXED(31) INIT (1(31));
   DCL LENBUF  FIXED(31);

   LENBUF = ROOT.Indexw;
 #FIN;

 IF ROOT IS NIL THEN RETURN; FIN;

 Tofree = ROOT.Next;

 WHILE Tofree ISNT NIL REPEAT
  Nxtblock = Tofree.Next;
  FREE(Tofree);
  #IF MAXILOG;
    RMCOUNT=RMCOUNT+1(31);
  #FIN;
  Tofree = Nxtblock;
 END;

 FREE(ROOT); ROOT = NIL;

 #IF MAXILOG;
  PUT 'S or Ax: (RM_LIST)',LENBUF,' bytes in ',RMCOUNT, ' blocks removed'
    TO loglink BY A,F(8),A,F(4),A;
 #FIN;

 RETURN;
END;  ! RM_LIST


! *********************************************************************
!      This procedure writes byte-data into a chained memory list
! *********************************************************************

WRITE_TO_LIST: PROC /*+G*/ ( ROOT REF LISTE IDENT,
                             Txt REF INV STRUCT[],
                             Txtlen FIXED(31),
                             loglink DATION OUT ALPHIC IDENT
                           ) GLOBAL;

 DCL IndexPTR REF FIXED(31);             ! For fast access to index
 DCL (Actblock,Newblock) REF LISTE;      ! For fast access to blocks
 DCL Blockno FIXED(31);                  ! Current block number
 DCL Bytes_to_go FIXED(31);              ! to be done in one block
 DCL Locindex FIXED(31);                 ! Index inside actual block
 DCL Space FIXED(31);                    ! Space inside actual block

 DCL (C1PTRS) REF INV CHAR(1);        ! For case of char(1) xfer
 DCL (C1PTRT) REF     CHAR(1);        ! For case of char(1) xfer
 DCL (C16PTRS) REF INV CHAR(16);     ! For case of char(16) xfer
 DCL (C16PTRT) REF     CHAR(16);     ! For case of char(16) xfer
 DCL (C32PTRS) REF INV CHAR(32);     ! For case of char(32) xfer
 DCL (C32PTRT) REF     CHAR(32);     ! For case of char(32) xfer
 DCL (C48PTRS) REF INV CHAR(48);     ! For case of char(48) xfer
 DCL (C48PTRT) REF     CHAR(48);     ! For case of char(48) xfer
 DCL (C64PTRS) REF INV CHAR(64);     ! For case of char(64) xfer
 DCL (C64PTRT) REF     CHAR(64);     ! For case of char(64) xfer
 DCL (C96PTRS) REF INV CHAR(96);     ! For case of char(96) xfer
 DCL (C96PTRT) REF     CHAR(96);     ! For case of char(96) xfer
 DCL (C128PTRS) REF INV CHAR(128);  ! For case of char(128) xfer
 DCL (C128PTRT) REF     CHAR(128);  ! For case of char(128) xfer
 DCL (C160PTRS) REF INV CHAR(160);  ! For case of char(160) xfer
 DCL (C160PTRT) REF     CHAR(160);  ! For case of char(160) xfer
 DCL (C192PTRS) REF INV CHAR(192);  ! For case of char(192) xfer
 DCL (C192PTRT) REF     CHAR(192);  ! For case of char(192) xfer

 DCL Target REF STRUCT[];                ! where to write

! .... First we have to check for a valid root-pointer .................

 IF ROOT IS NIL THEN
   ROOT = MALLOC(SIZEOF(ROOT),1);
   IF ROOT IS NIL THEN                   ! Did MALLOC succeed?
     #IF MAXILOG+MINILOG;
       WRITE 'S or Ax: Malloc failed, insuff. memory' TO loglink;
     #FIN;
     RETURN;                             ! Very bad: ignore request
   FIN;
   ROOT.Next = NIL; ROOT.Pred = NIL;     ! Init the links
   ROOT.Last = ROOT;                     ! Last block is root itself
   ROOT.Indexw  = 0(31) ;                ! Reset because this is a virgin
   ROOT.datptr  = NIL   ;
   ROOT.datlen  = 0(31) ;
   ROOT.headlen = 0(31) ;
 FIN;

 IndexPTR = ROOT.Indexw;                 ! The writer index (fast access)

! .... Determine the block to write to ................................

 Actblock = ROOT.Last;
 Locindex=TOFIXED(TOBIT(IndexPTR FIT 1(15)) AND TOBIT(BLOCKSIZE -1));  ! extract
 CONT IndexPTR = IndexPTR+Txtlen;

! .... Deposit the incoming text inside the list ......................
! .... Due to the fact, that the string-move generated by the compiler
! .... is much faster than any byte by byte copy, we use variable-length
! .... pointers. If the "BY TYPE" way is accepted: rearrange this!

 WHILE Txtlen > 0 OR Locindex == BLOCKSIZE REPEAT ! As long as there are data
   Space = BLOCKSIZE-Locindex;
   IF Space LE 0(31) THEN                ! New block required ?
      Newblock = MALLOC(SIZEOF(ROOT),1); ! Grab a new block
      IF Newblock IS NIL THEN
        #IF MAXILOG+MINILOG;
          WRITE 'S or Ax: Malloc failed, insuff. memory' TO loglink;
        #FIN;
        CONT IndexPTR = IndexPTR-Txtlen; ! Recorrect fill state
        RETURN;                          ! Very bad: ignore request
      FIN;
      Newblock.Next = NIL;               ! Is the last
      Newblock.Pred = Actblock;          ! Backlink
      Actblock.Next = Newblock;          ! Forw. link of old
      Actblock = Newblock;               ! To deal with
      ROOT.Last= Actblock;               ! For later access
      Locindex = 0(31);                  ! where to store data
      Space = BLOCKSIZE;                 ! full size of block available
   FIN;

   Target = Actblock.Array(Locindex);    ! The target address
   IF Space GE Txtlen THEN               ! all will fit
     Bytes_to_go = Txtlen;               ! Take the full incoming text
   ELSE
     Bytes_to_go = Space;                ! Take only a reduced 1st part
   FIN;

! .... Depending on the actual length a speed optimized xfer is selected

   CASE (Bytes_to_go FIT 1(15))//16

     ALT(0)     C1PTRS = Txt; C1PTRT = Target;         ! lame byte by byte
                TO Bytes_to_go REPEAT                  ! max 15
                   CONT C1PTRT = C1PTRS;
                   REFADD(C1PTRT,1);REFADD(C1PTRS,1);
                END;
                Locindex = Locindex+Bytes_to_go;
                Txtlen   = Txtlen-Bytes_to_go;
                Txt = C1PTRS;

     ALT(1)     C16PTRS = Txt; C16PTRT = Target;       ! 16 ... 31
                CONT C16PTRT = C16PTRS;
                Locindex = Locindex+16(31);
                Txtlen = Txtlen-16(31);
                REFADD(C16PTRS,1); Txt = C16PTRS;

     ALT(2)     C32PTRS = Txt; C32PTRT = Target;       ! 32 ... 47
                CONT C32PTRT = C32PTRS;
                Locindex = Locindex+32(31);
                Txtlen = Txtlen-32(31);
                REFADD(C32PTRS,1); Txt = C32PTRS;

     ALT(3)     C48PTRS = Txt; C48PTRT = Target;       ! 48 ... 63
                CONT C48PTRT = C48PTRS;
                Locindex = Locindex+48(31);
                Txtlen = Txtlen-48(31);
                REFADD(C48PTRS,1); Txt = C48PTRS;

     ALT(4,5)   C64PTRS = Txt; C64PTRT = Target;       ! 64 ... 95
                CONT C64PTRT = C64PTRS;
                Locindex = Locindex+64(31);
                Txtlen = Txtlen-64(31);
                REFADD(C64PTRS,1); Txt = C64PTRS;

     ALT(6,7)   C96PTRS = Txt; C96PTRT = Target;       ! 96 ... 127
                CONT C96PTRT = C96PTRS;
                Locindex = Locindex+96(31);
                Txtlen = Txtlen-96(31);
                REFADD(C96PTRS,1); Txt = C96PTRS;

     ALT(8,9)   C128PTRS = Txt; C128PTRT = Target;     ! 128 ... 159
                CONT C128PTRT = C128PTRS;
                Locindex = Locindex+128(31);
                Txtlen = Txtlen-128(31);
                REFADD(C128PTRS,1); Txt = C128PTRS;

     ALT(10,11) C160PTRS = Txt; C160PTRT = Target;     ! 160 ... 191
                CONT C160PTRT = C160PTRS;
                Locindex = Locindex+160(31);
                Txtlen = Txtlen-160(31);
                REFADD(C160PTRS,1); Txt = C160PTRS;

     OUT        C192PTRS = Txt; C192PTRT = Target;     ! 192 ... oo
                CONT C192PTRT = C192PTRS;
                Locindex = Locindex+192(31);
                Txtlen = Txtlen-192(31);
                REFADD(C192PTRS,1); Txt = C192PTRS;

   FIN;

 END;   ! while Textlen > 0

END;

! *********************************************************************
!      This procedure returns the IP-address and port of the remote client
! *********************************************************************

GET_HTTP_REMADDR: PROC /*+G*/ ( IPADDR FIXED(31) IDENT, PORT FIXED(15) IDENT) GLOBAL;

DCL USER FIXED;
DCL AUTP REF TUSER;

  USER = GET_USER;

  FOR I FROM MAX_USER BY -1 TO 1 REPEAT;
     AUTP = USERTAB(I);
     IF ( AUTP.USERNO == USER ) AND ( AUTP.IPADR /= 0 )
     THEN
       IPADDR = AUTP.IPADR;
       PORT = AUTP.RMPADR;
       RETURN;  ! if User found
     FIN;
   END; ! I from MAX_USER to 1

  IPADDR = USER; ! if error
  PORT = 0;

END;
